
var net = require('net');
const EventEmitter = require('events');

const STATE_NONE = 0
const STATE_CONNECTING = 1
const STATE_CONNECTED = 2
const STATE_DISCONNECTED = 3
const STATE_RETRYING = 4


const JRPCConnectTimeout = 3000;
const JRPCReconnectTime = 2000;
const JRPCTTLTimeout1 = 20000;
const JRPCTTLTimeout2 = 5000;


var indexOf = [].indexOf || function(item) { for (var i = 0, l = this.length; i < l; i++) { if (i in this && this[i] === item) return i; } return -1; };

let abstract = function(code, message) {
		if (code == null) {
			code = -32099;  
		}
		return {
			'code': code,
			'message': message
		};
	};

let	systemError = function() {
    	return abstract( -32010, 'SystemError' );
	};

let	networkError = function() {
		return abstract( -32098, 'NetworkError' );
	};

let	methodNotFound = function() {
		return abstract( -32601, 'MethodNotFound' );
	};

class JRPC2 extends EventEmitter {
	constructor() {
		super();
		const self = this;
		self.methods = 'methods';
		self.methodArgs = {};
		self.modules = {};
		self.context = {};
		self.buffer = '';
		self.socket = null;

		self.state = STATE_NONE;
		self.waitTimer = null;
		self.ttlTimer1 = null;
		self.ttlTimer2 = null;
	}

	_clearAllTimer() {
		const self = this;
		clearTimeout(self.waitTimer);
		clearTimeout(self.ttlTimer1);
		clearTimeout(self.ttlTimer2);
	}

	_startTTLTimer() {
		const self = this;
		self.ttlTimer1 = setTimeout(function(){
			//console.log('*** ttlTimer1 on');
			self.socket.write('\t\n');
			self.ttlTimer2 = setTimeout(function(){
				console.log('# Motebus.timeout(TTL).');
				self.socket.end();
				self.doDisconneced();
			}, JRPCTTLTimeout2);
		}, JRPCTTLTimeout1);
	}

	_resetTTLTimer() {
		const self = this;
		clearTimeout(self.ttlTimer1);
		clearTimeout(self.ttlTimer2);
		self._startTTLTimer();
	}

	_pauseTTLTimer() {
		const self = this;
		//console.log('*** ttlTimer1 off');
		clearTimeout(self.ttlTimer1);
		clearTimeout(self.ttlTimer2);
	}

	_initSocket() {
		const self = this;
		//console.log('*** _initSocket()')
		self.socket = new net.Socket(); 
		self.socket.setEncoding('utf8');
	}

	_connecting() {
		const self = this;
		if (self.state == STATE_CONNECTING) {
			//console.log('*** already _connecting()')
			return
		}
		//console.log('*** _connecting()')
		self.state = STATE_CONNECTING
		self._clearAllTimer();
		if (self.socket) {
			self.socket.removeAllListeners();
			self.socket.destroy()
			self.socket = null
		}

		self._initSocket();
		self.socket.connect(self.hostPort, self.hostIP, ()=>{
			self.socket.write('\t\n', ()=>{
				self.doConnected();
			});
		})
		self.socket.on('data', function(data) {
			self._resetTTLTimer();
			self.buffer += data.toString();
			//console.log( 'self.buffer = ', self.buffer );
			if (self.buffer.indexOf('\n') != -1) {
				var lines = self.buffer.split('\n');
				self.buffer = lines.pop();
				lines.forEach(function(line){
					if (line != "") {
						if (line == "\t") {
							self.socket.write('\b\n');
						} else if (line == "\b") {	
						} else
							self.doDispatch(line);
					}
				},this)
			}
		})
		.on('error', function(err) {
			//console.log('*** on error');
			clearTimeout(self.waitTimer);
			if ((err.code == "ECONNREFUSED") || (err.code == "ECONNRESET")) {
				console.log("!!! Please check that the %s program is running.", self.svrName)
			} else if (err.code == "ENETUNREACH") {
				console.log("!!! Please check network connectivity.", self.svrName)
			} else if (err.code == "EHOSTUNREACH") {
				console.log("!!! Please check setting, %s host unreachable.", self.svrName)
			} else if (err.code == "EISCONN") {
				// Do nothing
			} else {
				console.log('error: motebus.jrpc - ',err.code)
				//if (err.code == "ERR_STREAM_WRITE_AFTER_END") {}
			}
		})
		.on('close', function() {
			//console.log('*** on close');
			self.doDisconneced();
		})
		.on('end', function() {
			//console.log('*** on end');
			self.doDisconneced();
		});

		self.waitTimer = setTimeout(function(){
			//console.log('*** waitTimer active');
			self.socket.end();
		}, JRPCConnectTimeout);
	}

	_retrying() {
		const self = this;
		if (self.state == STATE_RETRYING) {
			//console.log('*** already _retrying()')
			return
		}
		//console.log('*** _retrying()')
		self.state = STATE_RETRYING

		setTimeout( ()=>{self._connecting()}, JRPCReconnectTime);	
	}

	connectTo( ip, port, svr ) {
		//console.log('*** call connectTo(%s:%d)', ip, port);
		const self = this;
		self.hostIP = ip;
		self.hostPort = port;
		self.svrName = svr;
		
		if (self.state == STATE_NONE) {
			setImmediate( ()=>{self._connecting()} );
		}
	}

	disconnect() {
		//console.log('*** call disconnect()')
		const self = this;
		if (self.state == STATE_CONNECTED) {
			self.socket.end();
			self.doDisconneced();
		}
	}


	doConnected() {
		const self = this
		if (self.state == STATE_CONNECTED) {
			//console.log('*** already doConnected()');	
			return
		}
		
		//console.log('*** doConnected()');
		self.state = STATE_CONNECTED
			
		clearTimeout(self.waitTimer);
		self.buffer = '';
		self.currid = 1;
		self.callbacks = {};
		self._startTTLTimer();
		self.emit('connect');
	}

	doDisconneced() {
		const self = this
		if (self.state == STATE_DISCONNECTED) {
			//console.log('*** already doDisconnected()')
			return
		}

		if (self.state == STATE_CONNECTED) {
			//console.log('*** doDisconneced()')
			self.state = STATE_DISCONNECTED
	
			self._clearAllTimer();
			this.emit('disconnect');
		}
		self._retrying()
	}

	doDispatch(msg) {
		//console.log('====> ' + msg);
		const self = this;
        var obj = JSON.parse(msg);

        if (Array.isArray(obj)) {
        	obj.forEach(function(item, index){
				if (item.method) {
					self.dispatchRequest(item);
				} else {
					self.dispatchResult(item);
				}
        	});
        } else {
			if (obj.method) {
				self.dispatchRequest(obj);
			} else {
				self.dispatchResult(obj);
			}
        }
	}

	
	splitMethod(methodName) {
		var method, res;
		if (indexOf.call(methodName, ".") >= 0) {
			res = methodName.split('.');
			method = res.pop();
			return [res.join('.'), method];
		} else {
			return [this.methods, methodName];
		}
	}

	getMethod(methodName) {
		var moduleName, ref;
		ref = this.splitMethod(methodName), moduleName = ref[0], methodName = ref[1];
		return this.modules[moduleName][methodName];
	}

	
	dispatchResult(obj) {
		//console.log('got response: ', obj);
		const self = this;
		//console.log('obj.id=',obj.id);
		var func = self.callbacks[obj.id];
		if (typeof func === 'function')
			func(obj.error, obj.result);
		delete self.callbacks[obj.id];
	}
	
	dispatchRequest(obj) {
		//console.log('incoming request: ', obj);
		const self = this;
		function responder(error, result) {
			var data;
			if (error) {
				data = {
					'jsonrpc': '2.0',
					'error': error,
					'id': obj.id
				};
			} else {
				data = {
					'jsonrpc': '2.0',
					'result': result,
					'id': obj.id
				};
			}
			var resp = JSON.stringify(data);

			if (self.state == STATE_CONNECTED) {
				//console.log('<==== ', resp);
				self.socket.write(resp+'\n');
			}
		}


		var len, j, name, args;
		var argNames = self.methodArgs[obj.method];
    	if (!argNames) {
			responder( methodNotFound(), null );
        	return;
      	}

		var func = self.getMethod(obj.method);
      	var result = null;

		var params = obj.params;
		if (params == null) {
			params = [];
		}

		try {
	      	if (params instanceof Array) {
	      		//console.log('params array');
	        	result = func.apply(self, params);
	      	} else {
	      		//console.log('params object');
	        	args = [];
	        	for (j = 0, len = argNames.length; j < len; j++) {
	          		name = argNames[j];
	          		args.push(params[name]);
	        	}
	        	result = func.apply(self, args); ////
	      	}

			if ((result != null) && typeof result.then === 'function') {
				result.then((res)=>{
					responder( null, res );
				}).catch((error)=>{
					if (error instanceof Error) {
						responder( abstract( null, error.message ), null );	
					} if ((error.code)&&(error.message)) {
						responder( error, null );
					} else {
						responder( abstract( null, error ), null );	
					}
				});
			} else {
				if (result instanceof Error) {
					//console.log('result normal error');
					responder( result, null );
				} else {
					//console.log('result normal ok');
					responder( null, result );
				}
			}
		} 
		catch(e) {
			console.log("Result Type - Error");

			var msg;
			if (e instanceof Error)
				msg = e.message
			else
				msg = e;
			responder( abstract( -32100, msg ), null );
		}
	}

	expose(fullName, func) {
		var args, methodName, moduleName, ref;

		args = func.toString().match(/function[^(]*\(([^)]*)\)/)[1];

		this.methodArgs[fullName] = args ? args.split(/,\s*/) : [];
		ref = this.splitMethod(fullName), moduleName = ref[0], methodName = ref[1];
		if (!this.modules[moduleName]) {
			this.modules[moduleName] = {};
		}
		//return this.modules[moduleName][methodName] = func;
		this.modules[moduleName][methodName] = func;
		return this;
	}


	call(method /*, params */) {
		const self = this;
		var id = self.currid;

		self.currid = self.currid + 1;
		if (self.currid > 16383) {self.currid = 1};

		var params = Array.prototype.slice.call(arguments);
		params.shift(); // drop `method`
		var req = JSON.stringify({
			'jsonrpc': '2.0',
			'method': method,
			'params': params,
			'id': id
		});

		return new Promise(function(resolve,reject){
			if (self.state == STATE_CONNECTED) {
				var cb = function(err,result) {
					//console.log('cb err=',err,' ,result=', result);
					if (err)
						reject(err);
					else
						resolve(result);
				};
				self.callbacks[id] = cb;
				
				//console.log('<==== ', req);
				setImmediate(function(){
					self._pauseTTLTimer();
					self.socket.write(req+'\n', ()=>{
						self._startTTLTimer();
					});
				});
			} else {
				reject(networkError());
			}
		});
	}

	notify(method /*, params */) {
		const self = this;

		var params = Array.prototype.slice.call(arguments);
		params.shift(); // drop `method`
		var req = JSON.stringify({
			'jsonrpc': '2.0',
			'method': method,
			'params': params
		});

		setImmediate(function(){
			self._pauseTTLTimer();
			self.socket.write(req+'\n', ()=>{
				self._startTTLTimer();
			});
		});
	}


	get isConnected() {
		const self = this;
		return (self.state == STATE_CONNECTED);
	}
}




module.exports.create = function() {
	return new JRPC2();	
}

module.exports.systemError = systemError;
